<?php

const QUIQQER_SYSTEM = true;
const QUIQQER_CONSOLE = true;

require_once dirname(__FILE__, 5) . '/header.php';

use PCSG\RabbitMQ\Channel\Channel;
use QUI\QueueManager\JobStatus;
use QUI\QueueManager\QueueJobUUID;
use QUI\QueueManager\QuiqqerWorkerQueueJob;
use QUI\QueueManager\Repository\QueueJobRepository;
use QUI\RabbitMQServer\Provider;

// run as daemon (forever)
set_time_limit(0);

$provider = new Provider();
$server = $provider->getQueueServer();

$jobRepository = new QueueJobRepository();

$Channel = null;
$CurrentWorker = null;
$currentPriority = 1;
$currentJobId = 0;
$minPriority = 1;
$exit = false;
$isProcessing = false;

if (!empty($argv[1])) {
    $minPriority = (int)$argv[1];
}

if ($minPriority < 1) {
    $minPriority = 1;
} elseif ($minPriority > 255) {
    $minPriority = 255;
}

// settings
$Conf = QUI::getPackage('quiqqer/rabbitmqserver')->getConfig();
$memoryThreshold = (int)$Conf->get('jobchecker', 'critical_memory_threshold');
$workerCount = (int)$Conf->get('consumer', 'consumer_count');
$debug = !empty($Conf->get('server', 'debug'));

/**
 * Check memory usage and exit if threshold is exceeded
 *
 * @return void
 */
function memoryCheck()
{
    global $memoryThreshold;
    global $workerCount;

    $currentMemoryUsage = (memory_get_usage() / 1024) / 1024;
    $approxTotalUsage = $currentMemoryUsage * $workerCount;

    if ($approxTotalUsage > $memoryThreshold) {
        QUI\System\Log::addInfo(
            'RabbitConsumer (pid: ' . getmypid() . ') exiting because total memory usage has'
            . ' reached ' . $approxTotalUsage . 'MB'
        );

        shutdown();
    }
}

/**
 * Requeue current Job
 *
 * @return void
 */
function requeue()
{
    global $CurrentWorker;
    global $currentPriority;
    global $currentJobId;

    /** @var \QUI\QueueManager\AbstractQuiqqerQueueWorker $CurrentWorker */
    if (empty($CurrentWorker)) {
        callbackEnd();
        return;
    }

    sleep(1);
    $CurrentWorker->cloneJob($currentPriority);

    QUI\System\Log::addInfo(
        'Re-queueing Job #' . $currentJobId . ' ("' . $CurrentWorker::class . '")'
    );

    callbackEnd();
}

/**
 * Checks if an exception is a Database Exception
 *
 * @param Exception $Exception
 * @return bool
 */
function isDBException(\Exception $Exception)
{
    if (mb_strpos($Exception->getMessage(), 'Error while sending QUERY packet') !== false) {
        return true;
    }

    return false;
}

$errorHandler = function () {
    global $CurrentWorker;

    $error = error_get_last();

    if (is_null($error)) {
        QUI\System\Log::addError(
            'RabbitConsumer shutdown'
        );
    } else {
        QUI\System\Log::addError(
            'RabbitConsumer shutdown :: ' . json_encode($error)
        );
    }

    /** @var \QUI\QueueManager\AbstractQuiqqerQueueWorker $CurrentWorker */
    if (empty($CurrentWorker)) {
        return;
    }

    requeue();

    QUI\System\Log::addDebug(
        'Cloning Job "' . $CurrentWorker::class . '" because of error -> ' . $error['type'] . ": " . $error['message']
    );
};

/**
 * Check the QUIQQER database connection and rebuilt
 * it if necessary
 *
 * @return void
 */
function checkDBConnection()
{
    try {
        QUI::getDataBase()->fetch([
            'count' => 1,
            'from' => QUI::getUsers()::table(),
            'limit' => 1
        ]);
    } catch (\Exception $Exception) {
        QUI\System\Log::addWarning(
            'MySQL connection seems to be lost, reconnecting. Database error that was thrown: '
            . $Exception->getMessage()
        );

        QUI::$DataBase2 = null;
    }
}

register_shutdown_function($errorHandler);

// execute job
function callbackEnd()
{
    global $Channel;
    global $CurrentWorker;
    global $exit;
    global $isProcessing;

    $CurrentWorker = null;
    $isProcessing = false;

    if ($exit) {
        $Channel->closeConnection();
        exit;
    }
}

$callback = function ($msg) {
    global $Channel;
    global $CurrentWorker;
    global $minPriority;
    global $currentPriority;
    global $currentJobId;
    global $isProcessing;
    global $debug;
    global $server;
    global $jobRepository;

    $isProcessing = true;
    $jobUUID = new QueueJobUUID($msg->body);
    $job = $jobRepository->find($jobUUID);

    if (is_null($job)) {
        QUI\System\Log::addError(
            'RabbitConsumer.php :: Job with UUID "' . $jobUUID . '" not found. Abort process.'
        );

        $Channel->ack($msg->delivery_info['delivery_tag']);
        callbackEnd();
        return;
    }

    if (!($job instanceof QuiqqerWorkerQueueJob)) {
        QUI\System\Log::addError(
            'RabbitConsumer.php :: Job with UUID "' . $jobUUID . '" is not a QUIQQER worker job. Abort process.'
        );

        $Channel->ack($msg->delivery_info['delivery_tag']);
        callbackEnd();
        return;
    }

    // Check priority
    if ($job->jobAttributes->priority < $minPriority) {
        $Channel->nack($msg->delivery_info['delivery_tag']);
        callbackEnd();
        return;
    } else {
        $Channel->ack($msg->delivery_info['delivery_tag']);
    }

    $currentPriority = $job->jobAttributes->priority;

    // Decode and execute job
    checkDBConnection();

    try {
        $currentJobId = $job->uuid->uuid;
        $workerClass = $job->getQuiqqerQueueWorkerClass();

        /** @var \QUI\QueueManager\AbstractQuiqqerQueueWorker $Worker */
        $Worker = new $workerClass();
        $CurrentWorker = $Worker;
    } catch (\Exception $Exception) {
        if (isDBException($Exception)) {
            requeue();
            return;
        }

        QUI\System\Log::addError(
            'RabbitConsumer.php (Job #' . $job->uuid . ') :: Error while initializing Worker class'
            . ' (' . $job->getQuiqqerQueueWorkerClass() . ') -> '
            . $Exception->getMessage() . '. Abort process.'
        );

        QUI\System\Log::writeException($Exception);

        callbackEnd();
        return;
    }

    try {
        $server->setJobStatus($job->uuid, JobStatus::RUNNING);

        $server->writeDebugMsg('Starting Queue Job #' . $job->uuid);
        $server->setJobResult($job->uuid, $Worker->execute($job));
        $server->writeDebugMsg('Finish Queue Job #' . $job->uuid);

        $server->setJobStatus($job->uuid, JobStatus::FINISHED);
        $server->writeDebugMsg('Status set to FINISHED for Job #' . $job->uuid);

        unset($Worker);

        // start garbage collection manually to try to clean up any memory leakage
        gc_collect_cycles();
    } catch (\Exception $Exception) {
        if (isDBException($Exception)) {
            requeue();
            return;
        }

        QUI\System\Log::addError(
            'RabbitConsumer.php (Job #' . $job->uuid . ') :: Error while executing Worker'
            . ' (' . $job->getQuiqqerQueueWorkerClass() . ') -> '
            . $Exception->getMessage() . '. Abort process.'
        );

        QUI\System\Log::writeException($Exception);

        $server->setJobStatus($job->uuid, JobStatus::ERROR);

        callbackEnd();
        return;
    }

    if ($job->jobAttributes->deleteOnFinish) {
        try {
            $server->deleteJob($job->uuid);
        } catch (\Exception $Exception) {
            if (isDBException($Exception)) {
                requeue();
                return;
            }

            QUI\System\Log::addError(
                'RabbitConsumer.php (Job #' . $job->uuid . ') :: Error while deleteting Job -> '
                . $Exception->getMessage() . '. Abort process.'
            );

            QUI\System\Log::writeException($Exception);

            $server->setJobStatus($job->uuid, JobStatus::ERROR);
        }
    }

    // Check memory usage
    memoryCheck();

    // end callback
    callbackEnd();
};

/**
 * Shutdown process
 *
 * @return void
 */
function shutdown()
{
    global $Channel;
    global $exit;
    global $isProcessing;

    if (!$isProcessing) {
        if ($Channel) {
            $Channel->closeConnection();
        }

        exit;
    }

    $exit = true;
}

declare(ticks=1);

pcntl_signal(SIGINT, "shutdown");
pcntl_signal(SIGTERM, "shutdown");
pcntl_signal(SIGHUP, "shutdown");

/**
 * Get Channel to RabbitMQ server
 *
 * @return Channel
 */
function getChannel(): Channel
{
    global $server;

    try {
        $Channel = $server->getChannel(true);
    } catch (\Exception $Exception) {
        QUI\System\Log::addError(
            'Exiting RabbitConsumer because there was a connection error to the RabbitMQ server'
            . ': ' . $Exception->getMessage()
        );

        exit;
    }

    return $Channel;
}

//$Channel->queue_declare('quiqqer_queue', false, true, false, false);
//$Channel->queue_bind('quiqqer_queue', 'quiqqer_exchange');
$Channel = getChannel();
$Channel->registerConsumerCallback($server->getUniqueQuiqqerQueueName(), $callback);

while (count($Channel->getAMQPChannel()->callbacks)) {
    try {
        if (!$Channel->isConnected()) {
            throw new Exception('Channel or Connection is null, fetch new connection');
        }

        $Channel->getAMQPChannel()->wait();
    } catch (\Exception $Exception) {
        $Channel = getChannel();
        $Channel->registerConsumerCallback($server->getUniqueQuiqqerQueueName(), $callback);
    }
};
